/*  rectify-jpg.c				23 May 2019           by Rich Griffiths  W2RG
 * 					revised		29 June 2019
 */

/*  under Linux, compile using:  gcc rectify-jpg.c -lm -ljpeg -o rectify-jpg
 *
 * rectify-jpg reads the jpg file and extracts one line of RGBRGBRGB... pixels at a time
 * using the IJG libjpeg library.
 * The pixels are uniformly spaced, which is an error referred to as tangential scale distortion.
 * An additional source of distortion is due to the curvature of Earth.
 * rectify-jpg calculates what the correct spacing of the pixels should be based on how the satellite's
 * scanner operates and the geometry of the orbit.
 * rectify-jpg then constructs a new line with properly spaced RGB pixels, using interpolated pixel
 * values to fill the gaps.  It is not strictly necessary to calculate the spacings every time
 * the program is run, but I (RAGS), having labored long and hard over the calculate spacings routine,
 * chose to leave that part of the code in place and accept the small inefficiency of doing so.
 * Finally, rectify-jpg then writes the rectified line to a new file using the IJG library.
 */

#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>
#include    <ctype.h>
#include    <math.h>
#include	<stddef.h>
#include	"jpeglib.h"
#include	<setjmp.h>

#define		QUALITY				100						// jpg image quality
#define		PI					3.1415926539			// one of my favorite desserts
#define		ALTITUDE			850						// altitude of satellite in kilometers
#define		REARTH				6357					// radius of Earth in kilometers
#define     SCAN_ANGLE			108.0					// Meteor's full scan angle in degrees
#define		PHI_MAX        		SCAN_ANGLE*PI/360.0		// half the scan angle, in radians
#define		BETA_MAX			PHI_MAX * ALTITUDE / REARTH		// longitude angle, center to west edge

int		validate_commandline ( int );
void	create_outfilename ( char *, char * );
void	calculate_pixel_spacing ( JSAMPARRAY, int, int * );

void	rectify_gray ( JSAMPARRAY, int, JSAMPARRAY, int );
void	rectify_RGB ( JSAMPARRAY, int, JSAMPARRAY, int );


int main( int argc, char** argv )  {

	char *infilename;
	char outfilename[ 128 ];
	FILE * fpin;
	FILE * fpout;

	JSAMPARRAY in_buffer;							// pointer to buffer for one line of pixels
	int in_buffer_width;							// actual number of bytes in in_buffer
	JSAMPARRAY rectifiedline;						// buffer for one rectified line of pixels
	int rectifiedline_width;
	
	/*  set up the input and output files   ****************************************/

	if ( validate_commandline ( argc ) )    {
		printf("\nrectify-jpg - corrects for tangential scale distortion in a jpg of a Meteor scan.\n\n");
		printf("Usage:\t rectify-jpg somefilename.jpg\n\n");
		exit ( -1 );
	}
                
    infilename = argv[1];
	if ( ( fpin = fopen( infilename, "rb" ) ) == NULL )	{
		fprintf( stderr,"\n\nCan't open file %s.  Check for typos, correct path, etc.\n\n", infilename );
		exit ( -1 );
	}

	create_outfilename ( infilename, outfilename );
	if ( ( fpout = fopen( outfilename, "wb" ) ) == NULL )	{
		fprintf( stderr,"\n\nCan't open file %s.  Check for typos, correct path, etc.\n\n", outfilename );
		exit ( -1 );
	}

	/* set up an extended error handler   *******************************************/

	struct my_error_mgr {							// the extended error handler struct:
		struct jpeg_error_mgr pub;					// "public" fields
		jmp_buf setjmp_buffer;						// for return to caller */
	};

	typedef struct my_error_mgr * my_error_ptr;

	/*  create JPEG decompression object  ******************************************/

	struct jpeg_decompress_struct in_cinfo;
	struct my_error_mgr jerr;

	in_cinfo.err = jpeg_std_error( &jerr.pub );

	if ( setjmp(jerr.setjmp_buffer ) ) {
		jpeg_destroy_decompress( &in_cinfo );
		fclose( fpin );
		return 0;
	}
	jpeg_create_decompress( &in_cinfo );			// initialize the JPEG decompression object
	jpeg_stdio_src( &in_cinfo, fpin );				// specify the data source

	jpeg_read_header( &in_cinfo, TRUE );			// read the file parameters (header)

	jpeg_start_decompress( &in_cinfo );

	in_buffer_width = in_cinfo.output_width * in_cinfo.output_components;
	in_buffer = ( *in_cinfo.mem->alloc_sarray ) ( (j_common_ptr) &in_cinfo, JPOOL_IMAGE, in_buffer_width, 1 );

	/*  create JPEG compression object  ******************************************/

	struct jpeg_compress_struct out_cinfo;

	struct jpeg_error_mgr out_jerr;

	out_cinfo.err = jpeg_std_error( &out_jerr );

	jpeg_create_compress( &out_cinfo );

	jpeg_stdio_dest( &out_cinfo, fpout );

	out_cinfo.image_width = ceil( in_cinfo.image_width * tan( PHI_MAX ) * ( 1.0 + tan( BETA_MAX ) ) );
	out_cinfo.image_height = in_cinfo.image_height;			// image width and height, in pixels
	out_cinfo.input_components = in_cinfo.num_components;	// # of color components (bytes) per pixel
	out_cinfo.in_color_space = in_cinfo.out_color_space; 	// colorspace of input image
	jpeg_set_defaults( &out_cinfo );

	jpeg_set_quality( &out_cinfo, QUALITY, TRUE );

	jpeg_start_compress( &out_cinfo, TRUE );

	/*  create a buffer to hold the rectified scan line  *********************************/

	rectifiedline_width = out_cinfo.image_width * out_cinfo.input_components;
	rectifiedline = ( *out_cinfo.mem->alloc_sarray ) ( (j_common_ptr) &in_cinfo, JPOOL_IMAGE, rectifiedline_width, 1 );

	/*  read lines of pixels from fpin, rectify each line, and write it to fpout  *******/

	for ( int linenum = 0; linenum < in_cinfo.image_height; linenum++ )	{

		jpeg_read_scanlines( &in_cinfo, in_buffer, 1 );

		switch ( in_cinfo.num_components )	{		// different methods for grayscale and RGB
			case 1:
				rectify_gray( in_buffer, in_cinfo.image_width, rectifiedline, out_cinfo.image_width );
				break;
			case 3:
				rectify_RGB( in_buffer, in_cinfo.image_width, rectifiedline, out_cinfo.image_width );
				break;
			default:
				printf("Image is neither grayscale nor RGB.  Exiting ....\n");
				exit( -1 );
				break;
		}
		jpeg_write_scanlines( &out_cinfo, rectifiedline, 1 );
	}

	jpeg_finish_decompress( &in_cinfo );			// finish decompression
	jpeg_destroy_decompress( &in_cinfo );

	jpeg_finish_compress( &out_cinfo );
	jpeg_destroy_compress( &out_cinfo );

	fclose( fpout );

	exit ( 0 );
}



int validate_commandline ( int n_args ) {
        
	if ( n_args == 2 )      
		return ( 0 );
	else
		return ( -1 );
}


/*  create the outfile name using the infile name  ***************************/
void create_outfilename ( char * infilename, char * outfilename )	{
        
	int i, k;
	int namelen = strlen ( infilename );
	int dotpos = namelen - 3;
	char ext[] = "jpg";
        
	for ( i = 0; i < 128; i++ )						// make sure the current outfilename is blank
		outfilename[ i ] = '\0';
	for ( dotpos = namelen; infilename[ dotpos ] != '.'; dotpos-- ) // find the dot before the ext
		;

	strncpy ( outfilename, infilename, dotpos );
	strcat ( outfilename, "-rectified." );
	strcat ( outfilename, ext );
        
	return;
}



void rectify_gray ( JSAMPARRAY in_line, int n_pixels, JSAMPARRAY rect_line, int n_rect_pixels )	{
	
	JSAMPLE byteA, byteB;

	int in_line_ptr = 0;
	int rect_ptr = 0;					// pointer to location within the rectified buffer
	int gap[ n_pixels ];

	calculate_pixel_spacing ( in_line, n_pixels, gap );

	byteA = *( *in_line + in_line_ptr++ );			// get the leftmost pixel

	for ( int i = 0; i < gap[0]; i++ )
		*( *rect_line + rect_ptr++ ) = byteA;		// pad left end of the rectified line
	*( *rect_line + rect_ptr++ ) = byteA;			// now add the pixel itself

	for ( int i = 1; i < n_pixels; i++ )	{

		byteB = *( *in_line + in_line_ptr++ );	// get the next pixel

		switch ( *( gap + i - 1 ) )	{					// fill the gap between the two pixels
			case 0:
				break;
			case 1:
				*( *rect_line + rect_ptr++ ) = ( byteA + byteB ) / 2;		// pad the space
				break;
			case 2:
				*( *rect_line + rect_ptr++ ) = ( byteA + byteA + byteB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteA + byteA + byteB ) / 3;
				break;
			case 3:
				*( *rect_line + rect_ptr++ ) = ( byteA + byteA + byteB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteA + byteB ) / 2;
				*( *rect_line + rect_ptr++ ) = ( byteA + byteB + byteB ) / 3;
				break;
			default:
				printf("\nA problem occurred spacing the pixels.  Exiting ....\n" );
				exit(0);
				break;
		}      // end switch
		*( *rect_line + rect_ptr++ ) = byteB;		// and write the pixel
		byteA = byteB;
	}   // end for(i=1 ...

	while (rect_ptr < n_rect_pixels )				// pad right end of rectified line
		*( *rect_line + rect_ptr++ ) = byteB;
}



void rectify_RGB ( JSAMPARRAY in_line, int n_pixels, JSAMPARRAY rect_line, int n_rect_pixels )	{
	
	JSAMPLE byteRA, byteGA, byteBA;
	JSAMPLE byteRB, byteGB, byteBB;

	int in_line_ptr = 0;
	int rect_ptr = 0;							// pointer to location within the rectified buffer
	int gap[ n_pixels ];

	calculate_pixel_spacing ( in_line, n_pixels, gap );

	byteRA = *( *in_line + in_line_ptr++ );				// get the leftmost pixel
	byteGA = *( *in_line + in_line_ptr++ );
	byteBA = *( *in_line + in_line_ptr++ );

	for ( int i = 0; i < gap[0]; i++ )	{
		*( *rect_line + rect_ptr++ ) = byteRA;			// pad left end of the rectified line
		*( *rect_line + rect_ptr++ ) = byteGA;
		*( *rect_line + rect_ptr++ ) = byteBA;
	}

	*( *rect_line + rect_ptr++ ) = byteRA;				// now add the pixel itself
	*( *rect_line + rect_ptr++ ) = byteGA;
	*( *rect_line + rect_ptr++ ) = byteBA;

	for ( int pixelnum = 1; pixelnum < n_pixels; pixelnum++ )    {

		byteRB = *( *in_line + in_line_ptr++ );			// get another pixel
		byteGB = *( *in_line + in_line_ptr++ );
		byteBB = *( *in_line + in_line_ptr++ );

		switch ( *( gap + pixelnum - 1 ) )	{				// fill the gap between the two pixels
			case 0:
				break;
			case 1:
				*( *rect_line + rect_ptr++ ) = ( byteRA + byteRB ) / 2;		// pad the space
				*( *rect_line + rect_ptr++ ) = ( byteGA + byteGB ) / 2;
				*( *rect_line + rect_ptr++ ) = ( byteBA + byteBB ) / 2;
				break;
			case 2:
				*( *rect_line + rect_ptr++ ) = ( byteRA + byteRA + byteRB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteGA + byteGA + byteGB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteBA + byteBB + byteBB) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteRA + byteRB + byteRB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteGA + byteGB + byteGB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteBA + byteBB + byteBB) / 3;
				break;
			case 3:
				*( *rect_line + rect_ptr++ ) = ( byteRA + byteRA + byteRB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteGA + byteGA + byteGB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteBA + byteBB + byteBB) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteRA + byteRB ) / 2;
				*( *rect_line + rect_ptr++ ) = ( byteGA + byteGB ) / 2;
				*( *rect_line + rect_ptr++ ) = ( byteBA + byteBB ) / 2;
				*( *rect_line + rect_ptr++ ) = ( byteRA + byteRB + byteRB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteGA + byteGB + byteGB ) / 3;
				*( *rect_line + rect_ptr++ ) = ( byteBA + byteBB + byteBB) / 3;
				break;
			default:
				printf("\nA problem occurred spacing the pixels.  Exiting ....\n" );
				exit(0);
				break;
		}      // end switch
		
		*( *rect_line + rect_ptr++ ) = byteRB;		// and write the pixel
		*( *rect_line + rect_ptr++ ) = byteGB;
		*( *rect_line + rect_ptr++ ) = byteBB;
		byteRA = byteRB;
		byteGA = byteGB;
		byteBA = byteBB;
	}   // end for( int i ...

	while (rect_ptr < 3 * n_rect_pixels )	{			// and pad the end of the line
		*( *rect_line + rect_ptr++ ) = byteRB;
		*( *rect_line + rect_ptr++ ) = byteGB;
		*( *rect_line + rect_ptr++ ) = byteBB;
	}
}



void calculate_pixel_spacing ( JSAMPARRAY buffer, int n_pixels, int *gap )	{

	/*   a little geometry of the satellite, Earth,  and the scans   **********************/

	double phi;									// instantaneous scan angle from vertical
	double delta_phi;							// incremental scan angle pixel-to-pixel
	int halfwidth;								// half the number of pixels
	double center;
	double beta;
	double delta_beta;
	
	double unusedspace = 0.0;					// unfilled gaps between true pixel locations

	phi = -PHI_MAX;
	delta_phi = 2.0 * PHI_MAX / ( n_pixels - 1 );
	beta = BETA_MAX;
	delta_beta = 2.0 * BETA_MAX / ( n_pixels - 1 );
	
	halfwidth = n_pixels / 2.0;
	center = (double)halfwidth - 0.5;			// there's no pixel directly under the satellite

	/*  calculate the correct position of each pixel  ******************************/

	double newposition[ n_pixels ];

	for ( int i = 0; i < n_pixels; i++ )	{
		phi = ( (double)( i ) - center ) * delta_phi;
		beta = ( (double)( i ) - center ) * delta_beta;
		newposition[ i ] = center *  tan( phi ) * ( 1.0 + fabs( tan( beta ) ) );
	}

	unusedspace = 0.0;
	gap[ halfwidth - 1 ] = 0;
	for ( int i = halfwidth - 2; i >= 0; i-- )	{
		unusedspace += fabs( newposition[ i ] - newposition[ i + 1 ] ) - 1.0;
		if ( unusedspace >= 3.0 )	{
			gap[ i ] = 3;
			unusedspace -= 3.0;
		}
		else if ( unusedspace >= 2.0 )	{
			gap[ i ] = 2;
			unusedspace -= 2.0;
		}
		else if ( unusedspace >= 1.0 )	{
			gap[ i ] = 1;
			unusedspace -= 1.0;
		}
		else
			gap[ i ] = 0;
		
	}
	unusedspace = 0.0;
	gap[ halfwidth ] = 0;
	for ( int i = halfwidth + 1; i < n_pixels; i++ )	{
		unusedspace += fabs( newposition[ i ] - newposition[ i - 1 ] ) - 1.0;
		if ( unusedspace >= 3.0 )	{
			gap[ i ] = 3;
			unusedspace -= 3.0;
		}
		else if ( unusedspace >= 2.0 )	{
			gap[ i ] = 2;
			unusedspace -= 2.0;
		}
		else if ( unusedspace >= 1.0 )	{
			gap[ i ] = 1;
			unusedspace -= 1.0;
		}
		else
			gap[ i ] = 0;
		
	}
}
